//******************************************************************
//* SERVO PROJECTS WITH THE PIC18LF4620 AND CP2102
//* WRITTEN BY FRED EADY FOR SERVO MAGAZINE
//* LAST UPDATED 01/13/2009
//* CHANGES/ADDITIONS
//******************************************************************b
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>

// Settings for PIC18F4620
__CONFIG(1,IESODIS & FCMDIS & HS);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0113);
//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void putch(char c);  
char recvchar(void);
char sendchar(char data);
char CharInQueue(void);
void lcd_init(void); 
void lcd_send_byte( char address, char n );
void lcd_send_nibble( char n ); 
void lcd_gotoxy( char x, char y); 
void scroll_SERVO(void);
//******************************************************************
//*	PIC18F4620 RAM Definitions
//******************************************************************
char bytein,window_bot;
unsigned int tmsecs1,imsecs1;
//******************************************************************
//*	PIC18F4620 ALIAS DEFINITIONS
//******************************************************************
#define	RELAY1	LATB0
#define	RELAY2	LATB1
#define	RELAY3	LATB2
#define	RELAY4	LATB3
#define	OFF		0
#define ON		1
//******************************************************************
//*	  LCD DEFINITIONS
//******************************************************************
#define databus	   	LATD
#define lcdcntrl   	LATE
#define E 	   		0x01
#define RW  		0x02 
#define RS			0x04 
#define clrRW		lcdcntrl &= ~RW
#define setRW		lcdcntrl |= RW
#define clrRS		lcdcntrl &= ~RS
#define setRS		lcdcntrl |= RS
#define clrE		lcdcntrl &= ~E
#define setE    	lcdcntrl |= E
				   
#define lcdcls	   lcd_send_byte(0,0x01)
#define	line1	   lcd_gotoxy(1,1)
#define	line2	   lcd_gotoxy(2,1)

char LCD_INIT_STRING[5] = {0x28,0x08,0x01,0x06,0x0E};

// LCD CHARACTER LOCATER  "12345678";
char lcdmsg_ready[]		= " READY! ";
char lcdmsg_clr[]  		= "        ";
char lcdmsg_servo[]	= "       SERVO SERVO SERVO        ";
///******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
#define BAUD	57600		//desired baud rate
#define FOSC	20000000	//oscillator frequency
#define DIVIDER ((unsigned int)(FOSC/(16 * BAUD) - 1))
//1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_RX_BUFFER_SIZE  256     
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_TX_BUFFER_SIZE  256      
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
char USART_RxBuf[USART_RX_BUFFER_SIZE],USART_TxBuf[USART_TX_BUFFER_SIZE];
char USART_TxHead,USART_TxTail,USART_RxHead,USART_RxTail;
#define esc	0x1B
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
							  	
#define		disable_TMR1int  	TMR1IE = 0
#define		enable_TMR1int	 	TMR1IE = 1
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0
#define mdelay1(msecdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								TIMER1ON;					\
								while(tmsecs1 < msecdelay);
						  	
//******************************************************************
//*	putch FUNCTION
//******************************************************************
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}

//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt TIMERS(void)
{
  char usart_data,tmphead,tmptail;

	if((TMR1IF && TMR1IE))
	{
		TMR1IF = 0;
		TMR1H = 0xEC;	
		TMR1L = 0x77;
	
		++tmsecs1;
	}

  if(RCIF)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
   USART_RxHead = tmphead;        // store new index 

   if ( tmphead == USART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   USART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( USART_TxHead != USART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
     USART_TxTail = tmptail;      // store new index  

    TXREG = USART_TxBuf[tmptail];  // start transmission
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }
} 
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
// load baud rate divisor for 576000 bps with 20MHz clock  
  SPBRG = DIVIDER;	
  TRISC7 = 1;		//receive pin
  TRISC6 = 0;		//transmit pin
  TXSTA = 0x04;		//high speed baud rate	
  RCSTA = 0x80;		//enable serial port and serial port pins
  
  USART_RxTail = 0x00;	//flush receive buffer 
  USART_RxHead = 0x00;
  USART_TxTail = 0x00;	//flush transmit buffer
  USART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable USART1 receiver
  TXIE = 0;			//disable USART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init(void)
{
 	char init8; 
	
  	TRISA = 0b11111111;		
   	TRISB = 0b11111111;		
   	TRISC = 0b11111111;
   	TRISD = 0b00000000;
   	TRISE = 0b00000000;
//******************************************************************
//*	CONFIGURE EEPROM READ/WRITE
//******************************************************************
	EECON1 = 0b00000000;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON0 = 0b00000000;	//turn OFF A2D
	ADCON1 = 0b00001111;	//all digital
//******************************************************************
//*	INITIALIZE THE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	CONFIGURE AND START TIMER1
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	TIMER1OFF;
	T1CON = 0b00000000;
	TMR1H = 0xEC;
	TMR1L = 0x77;
	TIMER1ON;
//******************************************************************
//*	ACTIVATE INTERRUPTS
//******************************************************************
	enable_TMR1int;
    enable_GLOBALint;
//******************************************************************
//******************************************************************
//*	INITIALIZE LCD
//******************************************************************
		lcd_init();
		lcdcls;
		line1;
		for(init8=0;init8<8;++init8)
			lcd_send_byte(1,lcdmsg_ready[init8]);
}
//******************************************************************
//*	GET A CHARACTER FROM RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
                                // wait for incomming usart_data  
  while ( USART_RxHead == USART_RxTail );
                                // calculate buffer index  
  tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxTail = tmptail;        // store new index  

  return USART_RxBuf[tmptail];   // return usart_data  
}

//******************************************************************
//*	SEND A CHARACTER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == USART_TxTail );
                                // store usart_data in buffer 
  USART_TxBuf[tmphead] = usart_data;
  USART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(USART_RxHead != USART_RxTail);
}
//******************************************************************
//*	LCD Functions
//******************************************************************
void lcd_send_nibble( char n ) 
{
	databus &= 0x0F;
	databus |= (n & 0xF0);
	mdelay1(1);
	setE;
	mdelay1(1);
	clrE;
}

void lcd_send_byte( char address, char n ) 
{
      clrE;
      
	  switch (address)
	  {
	  		 case 0:
			 	  clrRS;
				  break;
			 case 1:
			 	  setRS;
				  break;
			 default:
			 	  setRS;
				  break;	
      }		
	  		mdelay1(1);
		    
      lcd_send_nibble (n);
      lcd_send_nibble (n << 4);
	  		mdelay1(1);

}

void lcd_init(void) 
{
    char j8;
    clrRS;
    clrE;
    clrRW;
	mdelay1(16);
	lcd_send_nibble(0x30);
	mdelay1(6);
	lcd_send_nibble(0x30);
	mdelay1(1);
	lcd_send_nibble(0x30);
	mdelay1(2);
    lcd_send_nibble(0x20);
	for(j8=0;j8<5;++j8)
	{
	mdelay1(2);
       lcd_send_byte(0,LCD_INIT_STRING[j8]);
	}   
}

void lcd_gotoxy( char x, char y) 
{
// where x = lcd row (1,2,3,4) and y = column (1 thru 20)
   char address;
   switch (x)
   {
   		  case 1:
		  	   address = 0;
			   break;
		  case 2:
		  	   address = 0x40;
			   break;
		  case 3:
		  	   address = 0x14;
			   break;
		  case 4:
		  	   address = 0x54;
			   break;
		  default:
		  	   address = 0;
   }   
   address += (y-1);
   lcd_send_byte(0,0x80|address);
}
//******************************************************************
//*	SCROLL SERVO MESSAGE
//******************************************************************
void scroll_SERVO(void)
{
	char g8;
	for(g8=window_bot;g8<window_bot + 8;++g8)
		lcd_send_byte(1,lcdmsg_servo[g8]);
	mdelay1(150);

	if(++window_bot == 23)
		window_bot = 0;
}
//******************************************************************
//*	MAIN FUNCTION
//******************************************************************
void main(void)  
{
  	init();				//init the EUSART
  	line2;
  	do{
	  	while(!(CharInQueue()));
	  	bytein = recvchar();
	  	sendchar(bytein);
	  	if(bytein != 0x0D)
	  		lcd_send_byte(1,bytein);
	  }while(bytein != 0x0D);
	  	
	window_bot = 0;
	do{
		line1;
		scroll_SERVO();
	}while(1);	
}
