//******************************************************************
//* SERVO PROJECTS WITH THE PIC18LF4620 AND CP2102
//* WRITTEN BY FRED EADY FOR SERVO MAGAZINE
//* LAST UPDATED 01/09/2009
//* CHANGES/ADDITIONS
//******************************************************************b
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>

// Settings for PIC18F4620
__CONFIG(1,IESODIS & FCMDIS & HS);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0106);
//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void putch(char c); 
char recvchar(void);
char sendchar(char data);
char CharInQueue(void);
//******************************************************************
//*	PIC18F4620 RAM Definitions
//******************************************************************
char bytein;
unsigned int tmsecs1,imsecs1;
//******************************************************************
//*	PIC18F4620 ALIAS DEFINITIONS
//******************************************************************
#define	RELAY1	LATB0
#define	RELAY2	LATB1
#define	RELAY3	LATB2
#define	RELAY4	LATB3
#define	OFF		0
#define ON		1

///******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
#define BAUD	57600		//desired baud rate
#define FOSC	20000000	//oscillator frequency
#define DIVIDER ((unsigned int)(FOSC/(16 * BAUD) - 1))
//1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_RX_BUFFER_SIZE  256     
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_TX_BUFFER_SIZE  256      
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
char USART_RxBuf[USART_RX_BUFFER_SIZE],USART_TxBuf[USART_TX_BUFFER_SIZE];
char USART_TxHead,USART_TxTail,USART_RxHead,USART_RxTail;
#define esc	0x1B
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
							  	
//******************************************************************
//*	putch FUNCTION
//******************************************************************
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}

//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt TIMERS(void)
{
  char usart_data,tmphead,tmptail;

  if(RCIF)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
   USART_RxHead = tmphead;        // store new index 

   if ( tmphead == USART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   USART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( USART_TxHead != USART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
     USART_TxTail = tmptail;      // store new index  

    TXREG = USART_TxBuf[tmptail];  // start transmission
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }
} 
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
// load baud rate divisor for 576000 bps with 20MHz clock  
  SPBRG = DIVIDER;	
  TRISC7 = 1;		//receive pin
  TRISC6 = 0;		//transmit pin
  TXSTA = 0x04;		//high speed baud rate	
  RCSTA = 0x80;		//enable serial port and serial port pins
  
  USART_RxTail = 0x00;	//flush receive buffer 
  USART_RxHead = 0x00;
  USART_TxTail = 0x00;	//flush transmit buffer
  USART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable USART1 receiver
  TXIE = 0;			//disable USART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init(void)
{
 	char init8; 
	
  	TRISA = 0b11111111;		
   	TRISB = 0b11110000;		
   	TRISC = 0b11111111;
   	TRISD = 0b11111111;
   	TRISE = 0b00000000;
//******************************************************************
//*	CONFIGURE EEPROM READ/WRITE
//******************************************************************
	EECON1 = 0b00000000;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON0 = 0b00000000;	//turn OFF A2D
	ADCON1 = 0b00001111;	//all digital
//******************************************************************
//*	INITIALIZE THE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	ACTIVATE INTERRUPTS
//******************************************************************
    enable_GLOBALint;
//******************************************************************
}
//******************************************************************
//*	GET A CHARACTER FROM RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
                                // wait for incomming usart_data  
  while ( USART_RxHead == USART_RxTail );
                                // calculate buffer index  
  tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxTail = tmptail;        // store new index  

  return USART_RxBuf[tmptail];   // return usart_data  
}

//******************************************************************
//*	SEND A CHARACTER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == USART_TxTail );
                                // store usart_data in buffer 
  USART_TxBuf[tmphead] = usart_data;
  USART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(USART_RxHead != USART_RxTail);
}

}//******************************************************************
//*	MAIN FUNCTION
//******************************************************************
void main(void)
{

  	init();				//init the EUSART
  	RELAY1 = OFF;
  	RELAY2 = OFF;
  	RELAY3 = OFF;
  	RELAY4 = OFF;
  	do{
	  	while(!(CharInQueue()));
	  	bytein = recvchar();
	  	switch(bytein)
	  	{
		  	case '1':
		  		RELAY1 = ON;
		  		printf("RELAY 1 IS ON.\r\n");
		  	break;
		  	case '2':
		  		RELAY2 = ON;
		  		printf("RELAY 2 IS ON.\r\n");
		  	break;
		  	case '3':
		  		RELAY3 = ON;
		  		printf("RELAY 3 IS ON.\r\n");
		  	break;
		  	case '4':
		  		RELAY4 = ON;
		  		printf("RELAY 4 IS ON.\r\n");
		  	break;
		  	case esc:
	  			while(!(CharInQueue()));
	  			bytein = recvchar();
	  			switch(bytein)
	  			{
		  			case '1':
			  		RELAY1 = OFF;
			  		printf("RELAY 1 IS OFF.\r\n");
				  	break;
				  	case '2':
				  		RELAY2 = OFF;
				  		printf("RELAY 2 IS OFF.\r\n");
				  	break;
				  	case '3':
				  		RELAY3 = OFF;
				  		printf("RELAY 3 IS OFF.\r\n");
				  	break;
				  	case '4':
				  		RELAY4 = OFF;
				  		printf("RELAY 4 IS OFF.\r\n");
				  	break;
				}
		  	break;
		}

	  }while(1);	
}
